(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2013-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: display_window.ml,v 1.11 2014-02-10 18:12:49 fclement Exp $ *)

(* The graphics window is split into sub-windows. Each sub-window is
 surrounded by a white margin. Let n be the dimension of the parameter.
 
 Parameter_graphics -> n x 1 or 1 x n sub-windows,
 Full_graphics -> n x 2 or 2 x n sub-windows.
 
 Parameter_graphics means that all components of the parameter are displayed.
 Hence, n sub-windows.
 
 Full_graphics means that all components for both the parameter and the
 gradient of the cost function are displayed.
 Hence, 2n sub-windows.
 
 Depending on the aspect ratio of the parameter, components sub-windows are
 either displayed along lines or columns:
 for Parameter_graphics,
   Portrait -> n x 1 sub-windows
   Landscape -> 1 x n sub-windows,
 for Full_graphics,
   Portrait -> n x 2 sub-windows
   Landscape -> 2 x n sub-windows. *)

module Nat = Unatural;;
module Prational = Upositive_rational;;
module Prat_range = Upositive_rational_range;;

let _0_ = Nat._0_
and _1_ = Nat._1_
and _2_ = Nat._2_
and _3_ = Nat._3_
and _4_ = Nat._4_
;;

open Upositive_rational_range_types;;

let dw_qual = Upervasives.qualify "Display_window";;

let dw_invalid_arg = Upervasives.invalid_arg_qual dw_qual;;

open Refinement_types;;
open Display_types;;

(* Compound graphics window. *)

let aspect (w, h) = if w <= h then Portrait else Landscape;;

let comps_direction sz =
  match aspect sz with
  | Portrait -> Horizontal_comps_dir
  | Landscape -> Vertical_comps_dir
;;

let same_direction dir comps_dir =
  match (dir, comps_dir) with
  | (Horizontal_dir, Horizontal_comps_dir)
  | (Vertical_dir, Vertical_comps_dir) -> true
  | (Horizontal_dir, Vertical_comps_dir)
  | (Vertical_dir, Horizontal_comps_dir) -> false
;;

(* See description of the contents of the graphics window in the header. *)
let coeffs n gm dir comps_dir =
  let b = same_direction dir comps_dir in
  match gm with
  | No_graphics -> assert false
  | Parameter_graphics -> if b then n else _1_
  | Full_graphics -> if b then n else _2_
;;

(* nn = alpha*(l + 2*m). *)
let compound_length n gm dir comps_dir m l =
  let two_ms = Nat.mul _2_ m in
  let alpha = coeffs n gm dir comps_dir in
  let sub_length = Nat.add l two_ms in
  Nat.mul alpha sub_length
;;

(* Solve compound formula with l <- r*l for r.
 r = (nn - alpha*2*m) / (alpha*l). *)
let split_ratio n gm dir comps_dir m l nn =
  let two_ms = Nat.mul _2_ m in
  let alpha = coeffs n gm dir comps_dir in
  let p = Nat.sub nn (Nat.mul alpha two_ms)
  and q = Nat.mul alpha l in
  Prational.of_nat p q
;;

let margin = Nat.of_int 10;;

let scaling_factor n gm (nnx_min, nny_min) (nnx_max, nny_max) (w, h) =
  let comps_dir = comps_direction (w, h) in
  let rgx =
    Prat_range.of_rat
      (split_ratio n gm Horizontal_dir comps_dir margin w nnx_min)
      (split_ratio n gm Horizontal_dir comps_dir margin w nnx_max)
  and rgy =
    Prat_range.of_rat
      (split_ratio n gm Vertical_dir comps_dir margin h nny_min)
      (split_ratio n gm Vertical_dir comps_dir margin h nny_max) in
  let rg = Prat_range.intersection rgx rgy in
  match (rgx, rgy, rg) with
  | (Empty, _, _) | (_, Empty, _) ->
      dw_invalid_arg "scaling_factor" "incompatible window size bounds"
  | (Nonempty rgx, Nonempty rgy, Empty) ->
      Prational.min rgx.right rgy.right
  | (Nonempty _, Nonempty _, Nonempty _) ->
      Prat_range.project rg Prational._1_
;;

(* Graphics_elements. *)

let init_graphics_elements =
  {
   comps_direction = Horizontal_comps_dir;
   scaling_factor = Prational._1_;
   display_image_size = (Nat._0_, Nat._0_);
   window_size = (Nat._0_, Nat._0_);
  }
;;

(* This should be an abstract reference with unique assignment, weakening it to
 mandatory assignment is necessary for testing purposes though... *)
let { Uaref.get = get_graphics_elements;
      set = gen_set_graphics_elements; } =
  Uaref.make_aref_mandatory "graphics elements" init_graphics_elements
;;

let set_graphics_elements n gm min_w_sz max_w_sz ((w, h) as sz) =
  let comps_dir = comps_direction (w, h) in
  let sf = scaling_factor n gm min_w_sz max_w_sz sz in
  let scale = Prational.scale sf in
  let ((nx, ny) as d_sz) = (scale w, scale h) in
  let nnx = compound_length n gm Horizontal_dir comps_dir margin nx
  and nny = compound_length n gm Vertical_dir comps_dir margin ny in
  let ge =
    {
     comps_direction = comps_dir;
     scaling_factor = sf;
     display_image_size = d_sz;
     window_size = (nnx, nny);
    } in
  gen_set_graphics_elements ge
;;

(* Display window setup. *)

let open_display (nnx, nny) =
  Graphics.open_graph "";
  Graphics.resize_window (Nat.to_int nnx) (Nat.to_int nny);
;;

let window_setup n gm min_w_sz max_w_sz sz =
  set_graphics_elements n gm min_w_sz max_w_sz sz;
  let { window_size = w_sz; _; } = get_graphics_elements () in
  open_display w_sz;
;;
